@@ -33,6 +33,7 @@
 #include <signal.h>
 #include <sys/types.h>
 #include <sys/stat.h>
+#include <sys/wait.h>
 #include <fcntl.h>
 #include <termios.h>
 #include <getopt.h>
@@ -62,6 +63,7 @@ static GMainLoop *main_loop = NULL;
 
 static int services = 0;
 static gboolean tty_needs_reinit = FALSE;
+static gboolean tty_open_allowed = TRUE;
 static int signal_pipe[2];
 
 #define TTY_RX_MTU 65535
@@ -74,9 +76,13 @@ int tty_init(int services, const gchar *root_path,
 	struct termios options;
 	int fd, err, arg;
 	glong flags;
+	GIOChannel *io = NULL;
 
 	tty_needs_reinit = TRUE;
 
+	if (!tty_open_allowed)
+		return -EACCES;
+
 	fd = open(devnode, O_RDWR | O_NOCTTY);
 	if (fd < 0)
 		return fd;
@@ -109,12 +115,18 @@ int tty_init(int services, const gchar *root_path,
 	server->rx_mtu = TTY_RX_MTU;
 	server->tx_mtu = TTY_TX_MTU;
 
-	err = obex_session_start(fd, server);
+	io = g_io_channel_unix_new(fd);
+	g_io_channel_set_close_on_unref(io, TRUE);
+
+	err = obex_session_start(io, server);
+	g_io_channel_unref(io);
+
 	if (err < 0) {
 		server_free(server);
 		goto failed;
-	} else
-		tty_needs_reinit = FALSE;
+	}
+
+	tty_needs_reinit = FALSE;
 
 	debug("Successfully opened %s", devnode);
 
@@ -122,7 +134,8 @@ int tty_init(int services, const gchar *root_path,
 
 failed:
 	error("tty_init(): %s (%d)", strerror(-err), -err);
-	close(fd);
+	if (io == NULL)
+		close(fd);
 	return err;
 }
 
@@ -133,13 +146,20 @@ void tty_closed(void)
 
 static void sig_term(int sig)
 {
+	info("Terminating due to signal %d", sig);
 	g_main_loop_quit(main_loop);
 }
 
+static void sig_debug(int sig)
+{
+	toggle_debug();
+}
+
 static gboolean option_detach = TRUE;
 static gboolean option_debug = FALSE;
 
 static gchar *option_root = NULL;
+static gchar *option_root_setup = NULL;
 static gchar *option_capability = NULL;
 static gchar *option_devnode = NULL;
 
@@ -158,6 +178,8 @@ static GOptionEntry options[] = {
 				"Enable debug information output" },
 	{ "root", 'r', 0, G_OPTION_ARG_STRING, &option_root,
 				"Specify root folder location", "PATH" },
+	{ "root-setup", 'S', 0, G_OPTION_ARG_STRING, &option_root_setup,
+				"Root folder setup script", "SCRIPT" },
 	{ "symlinks", 'l', 0, G_OPTION_ARG_NONE, &option_symlinks,
 				"Enable symlinks on root folder" },
 	{ "capability", 'c', 0, G_OPTION_ARG_STRING, &option_capability,
@@ -177,7 +199,7 @@ static GOptionEntry options[] = {
 	{ NULL },
 };
 
-static void sig_usr1(int sig)
+static void sig_tty(int sig)
 {
 	if (write(signal_pipe[1], &sig, sizeof(sig)) != sizeof(sig))
 		error("unable to write to signal pipe");
@@ -189,13 +211,27 @@ static gboolean handle_signal(GIOChannel *io, GIOCondition cond,
 	int sig, fd = g_io_channel_unix_get_fd(io);
 
 	if (read(fd, &sig, sizeof(sig)) != sizeof(sig)) {
-		error("handle_sigusr1: unable to read signal from pipe");
+		error("handle_signal: unable to read signal from pipe");
 		return TRUE;
 	}
 
-	if (sig == SIGUSR1 && tty_needs_reinit)
-		tty_init(services, option_root, option_capability,
-				option_devnode);
+	switch (sig) {
+	case SIGUSR1:
+		debug("SIGUSR1");
+		tty_open_allowed = TRUE;
+		if (tty_needs_reinit)
+			tty_init(services, option_root, option_capability,
+							option_devnode);
+		break;
+	case SIGHUP:
+		debug("SIGHUP");
+		tty_open_allowed = FALSE;
+		obex_tty_session_stop();
+		break;
+	default:
+		error("handle_signal: got unexpected signal %d", sig);
+		break;
+	}
 
 	return TRUE;
 }
@@ -213,18 +249,56 @@ static int devnode_setup(void)
 	g_io_channel_unref(pipe_io);
 
 	memset(&sa, 0, sizeof(sa));
-	sa.sa_handler = sig_usr1;
+	sa.sa_handler = sig_tty;
 	sigaction(SIGUSR1, &sa, NULL);
+	sigaction(SIGHUP, &sa, NULL);
 
-	if (option_pcsuite) {
-		tty_needs_reinit = TRUE;
-		return 0;
-	}
+	if (option_pcsuite)
+		tty_open_allowed = FALSE;
 
 	return tty_init(services, option_root, option_capability,
 			option_devnode);
 }
 
+static gboolean is_dir(const char *dir) {
+	struct stat st;
+
+	if (stat(dir, &st) < 0) {
+		error("stat(%s): %s (%d)", dir, strerror(errno), errno);
+		return FALSE;
+	}
+
+	return S_ISDIR(st.st_mode);
+}
+
+static gboolean root_folder_setup(char *root, char *root_setup)
+{
+	gint status;
+	char *argv[3] = { root_setup, root, NULL };
+
+	if (is_dir(root))
+		return TRUE;
+
+	if (root_setup == NULL)
+		return FALSE;
+
+	debug("Setting up %s using %s", root, root_setup);
+
+	if (!g_spawn_sync(NULL, argv, NULL, 0, NULL, NULL, NULL, NULL,
+							&status, NULL)) {
+		error("Unable to execute %s", root_setup);
+		return FALSE;
+	}
+
+	if (WEXITSTATUS(status) != EXIT_SUCCESS) {
+		error("%s exited with status %d", root_setup,
+							WEXITSTATUS(status));
+		return FALSE;
+	}
+
+	return is_dir(root);
+}
+
 int main(int argc, char *argv[])
 {
 	GOptionContext *context;
@@ -292,6 +366,14 @@ int main(int argc, char *argv[])
 	if (option_root == NULL)
 		option_root = g_strdup(DEFAULT_ROOT_PATH);
 
+	if (option_root[0] != '/') {
+		char *old_root = option_root, *home = getenv("HOME");
+		if (home) {
+			option_root = g_strdup_printf("%s/%s", home, old_root);
+			g_free(old_root);
+		}
+	}
+
 	if (option_capability == NULL)
 		option_capability = g_strdup(DEFAULT_CAP_FILE);
 
@@ -326,11 +408,19 @@ int main(int argc, char *argv[])
 	if (option_devnode)
 		devnode_setup();
 
+	if (!root_folder_setup(option_root, option_root_setup)) {
+		error("Unable to setup root folder %s", option_root);
+		exit(EXIT_FAILURE);
+	}
+
 	memset(&sa, 0, sizeof(sa));
 	sa.sa_handler = sig_term;
 	sigaction(SIGINT, &sa, NULL);
 	sigaction(SIGTERM, &sa, NULL);
 
+	sa.sa_handler = sig_debug;
+	sigaction(SIGUSR2, &sa, NULL);
+
 	g_main_loop_run(main_loop);
 
 	bluetooth_exit();
